#include "gnuboy.h"
#include <kos.h>
#include <arch/gdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdarg.h>
#include <signal.h>

#include "input.h"
#include "mem.h"

#include "Version"
#include "dc/dc_utils.h"
#include "dc/dc_menu.h"
#include "dc/dc_vmu.h"

/* 
extern uint8 romdisk[];

KOS_INIT_ROMDISK(romdisk);
 */

#define DC_FILES_ROOTPATH "/cd"

static uint16 startup_image[320 * 240];
static uint16 menu_image[320 * 240];
static uint16 credits_image[320 * 240];
static uint16 options_image[320 * 240];

uint8 dc_mvmu = 0;

/*
uint16 cont_a_keycode;
uint16 cont_b_keycode;

enum
{
	bx_aa,
	ba_ab,
	bx_ab,
	cont_scheme_pattern_num
} cont_scheme_pattern;

const struct
{
	int scheme;
	char *scheme_str;
} scheme_settings[] = {
	{	bx_aa,	"[GB_B = DC_X] [GB_A = DC_A]",	},
	{	ba_ab,	"[GB_B = DC_A] [GB_A = DC_B]",	},
	{	bx_ab,	"[GB_B = DC_X] [GB_A = DC_B]",	},
};



void set_cont_scheme (uint8 pattern)
{
	if (pattern < cont_scheme_pattern_num)
	  cont_scheme = pattern;
	else
	  cont_scheme = bx_aa;

	switch (cont_scheme)
	{
	  case bx_aa:
		cont_b_keycode = CONT_X;
		cont_a_keycode = CONT_A;
		break;
	  case ba_ab:
		cont_b_keycode = CONT_A;
		cont_a_keycode = CONT_B;
		break;
	  case bx_ab:
		cont_b_keycode = CONT_X;
		cont_a_keycode = CONT_B;
		break;
	  default:
		cont_b_keycode = CONT_X;
		cont_a_keycode = CONT_A;
		break;
	}
}

void inc_cont_scheme ()
{
	++cont_scheme;
	if (cont_scheme == cont_scheme_pattern_num)
	  cont_scheme = 0;

	set_cont_scheme (cont_scheme);
}

uint16 get_cont_a () { return cont_a_keycode; }
uint16 get_cont_b () { return cont_b_keycode; }
*/

/* vmu screen image */
#include "dc/vmu_icons/kos.h"


void doevents()
{
#if 0
  event_t ev;
  int st;
#endif

  ev_poll();

#if 0
  while (ev_getevent(&ev))
  {
    if (ev.type != EV_PRESS && ev.type != EV_RELEASE)
      continue;
    st = (ev.type != EV_RELEASE);
    rc_dokey(ev.code, st);
  }
#endif
}



#include "hw.h"


static void
run_game(const char *filename)
{
  dc_vid_empty ();

  exit_game_loop = 0;

  gb_vid_init ();
  pcm_init ();
  mem_init ();

  dc_print ("loading ROM");
  if (loader_init(filename))
  {
    emu_reset();
    emu_run ();

    pcm_clear_buffer();
    loader_unload ();
  }

  gb_vid_close();
  pcm_close();
}


static void
draw_main_menu ()
{
  display_rawimage (menu_image);
}


static char last_romfile[256];

static void
do_rom_menu_and_run ()
{
  fm_result_t result;
  char *p;
  char last_romfile_dir[256];
  fm_config_t fm_config;

  sprintf (last_romfile_dir, "%s", last_romfile);
  p = strrchr(last_romfile_dir, '/');
  *p = '\0';

  fm_config.vertical_wait = 80;
  fm_config.horizontal_wait = 80;
  fm_config.root_path = DC_FILES_ROOTPATH;
  if (do_file_menu (&result, last_romfile_dir, &fm_config))
  {
    if (strcmp(result.path, last_romfile))
      unlink_quickstate ();

    run_game (result.path);
    sprintf (last_romfile, "%s", result.path);
  }
}



static void
draw_main_menu_vmu_icon ()
{
  char vmu_screen[48 * 32];
  int vmu_free_blocks;
  char str[32];

  timer_spin_sleep (50); /* need wait */
  vmu_free_blocks = ndc_vmu_check_free_blocks (NULL, dc_mvmu);

  vmu_icon_clear (vmu_screen, 1);
  vmu_icon_draw_string (vmu_screen, 1, 1, "VMU FREE:");
  sprintf(str, "%d BLOCKS", vmu_free_blocks);
  vmu_icon_draw_string (vmu_screen, 3, 7, str);

  vmu_icon_draw_string (vmu_screen, 1, 20, "gnuboy/DC");
  vmu_icon_draw_string (vmu_screen, 3, 26, "1.0.3-0.6");

  vmu_set_icon (vmu_screen);

  timer_spin_sleep (200); /* need wait */
}

/*
static void
draw_main_option_menu_on_of (int pos_y, uint8 which)
{
  if (which)
    draw_string (260, pos_y, _white, _none, "[ON]");
  else
    draw_string (260, pos_y, _white, _none, "[OFF]");
}


static void
draw_main_option_menu ()
{
  //char str[128];
  //fm_config_t fm_config;

  display_rawimage (options_image);

  draw_string (50, 30, _yellow, _none, "L+R: Return to main menu");

  draw_string (30, 50, _white, _none, "START: Adjust GB A and B buttons");
  //draw_string (50, 60, _white, _none, get_cont_scheme_string());

  draw_string (30, 80, _white, _none, "DOWN:  VMU support");
  //draw_main_option_menu_on_of (90, is_use_vmu ());

  draw_string (30, 90, _white, _none, "RIGHT: Use bilinear filter");
  //draw_main_option_menu_on_of (110, is_bi_filter ());

  draw_string (30, 110, _white, _none, "Y: Adjust screen position");
  draw_string (30, 120, _white, _none, "R + B: Set default");
  draw_string (30, 130, _white, _none, "X: Save config to VMU");

  draw_string (30, 150, _white, _none, "analog: Adjust rom menu wait");
  //sprintf(str, "vertical %d / horizontal %d",
	  //fm_config.vertical_wait, fm_config.horizontal_wait);
  //draw_string (30, 170, _white, _none, str);
    dc_vid_flip (draw_type_fullscreen);
}

static void
do_main_options ()
{
  cont_cond_t cont;
  uint16 all_buttons =
    CONT_START |
    CONT_DPAD_UP |
    CONT_DPAD_DOWN |
    CONT_DPAD_LEFT |
    CONT_DPAD_RIGHT |
    CONT_A |
    CONT_B |
    CONT_Y |
    CONT_X;

  draw_main_option_menu ();
  dc_vid_flip_fill_renderer (draw_type_fullscreen);
  wait_until_release_buttons (dc_controller_addr, CONT_X);

  for (;;)
  {
    cont_get_cond (dc_controller_addr, &cont);

    if (cont.ltrig && cont.rtrig)
      break;

    if (!(cont.buttons & CONT_START))
      dc_settings.inc_cont_scheme ();

    if (!(cont.buttons & CONT_DPAD_DOWN))
      dc_settings.toggle_use_vmu ();

    if (!(cont.buttons & CONT_DPAD_LEFT))
      dc_settings.toggle_exsound ();

    if (!(cont.buttons & CONT_DPAD_RIGHT))
      dc_settings.toggle_bi_filter ();

    if (!(cont.buttons & CONT_Y))
      do_adjust_screen_rect_menu ();

    if (!(cont.buttons & CONT_B) & cont.rtrig)
      dc_settings.set_default ();

    if (!(cont.buttons & CONT_X))
      dc_settings.save_to_vmu(mvmu);

      // for rom menu options
      fm_config_t *config = &(dc_settings.fm_config);

      // analog up
      if (cont.joyy < 5)
      {
	    if (fm_config.vertical_wait > 0)
	      --fm_config.vertical_wait;
      }
      // analog down
      if (cont.joyy > 250)
      {
	    if (fm_config.vertical_wait < 200)
	      ++fm_config.vertical_wait;
      }
      // analog right
      if (cont.joyx > 250)
      {
	    if (fm_config.horizontal_wait < 200)
	      ++fm_config.horizontal_wait;
      }
      // analog left
      if (cont.joyx < 5)
      {
	    if (fm_config.horizontal_wait > 0)
	      --fm_config.horizontal_wait;
      }



    draw_main_option_menu ();
    dc_vid_flip (draw_type_fullscreen);

    if (is_buttons_press (&cont, all_buttons))
    {
      dc_vid_flip_fill_renderer (draw_type_fullscreen);
      wait_until_release_buttons (dc_controller_addr, all_buttons);
      //wait release trigger for do_adjust_screen_rect_menu
      wait_until_release_trigger (dc_controller_addr);
    }
  }
}
*/
static void
do_credits()
{
  int state = 0;
  int count = 0;
  cont_cond_t cont;

  for (;;)
  {
    cont_get_cond (dc_controller_addr, &cont);

    if (cont.rtrig && cont.ltrig)
      break;

    display_rawimage (credits_image);

    if (++count == 20)
    {
      count = 0;
      state ^= 1;
    }

    if (state)
      draw_string (10, 10, _white, _black, "pull L+R");

    dc_vid_flip (draw_type_fullscreen);
  }
}


static void
load_main_menu ()
{
  cont_cond_t cont;

  draw_main_menu_vmu_icon ();

  for (;;)
  {
    cont_get_cond (dc_controller_addr, &cont);

    if (cont.rtrig && cont.ltrig &&
	!(cont.buttons & CONT_Y) &&
	!(cont.buttons & CONT_START))
      break;

    if (!(cont.buttons & CONT_A))
    {
      do_rom_menu_and_run ();
      draw_main_menu_vmu_icon ();
    }
/*
    if (!(cont.buttons & CONT_X))
      do_main_options ();
*/
    if (!(cont.buttons & CONT_Y))
      do_credits ();

    if (!(cont.buttons & CONT_B))
    {
      do_vmu_menu (dc_mvmu);
      draw_main_menu_vmu_icon ();
    }

    draw_main_menu ();
    dc_vid_flip (draw_type_fullscreen);
  }
}


static void
draw_startup_image (const char *string)
{
  display_rawimage (startup_image);
#ifdef __DC_PAL__
  draw_string(10, 200, _white, _black, "gnuboy/DC 1.0.3-0.6-PAL");
#else
  draw_string(10, 200, _white, _black, "gnuboy/DC 1.0.3-0.6");
#endif
  draw_string(10, 210, _white, _none, string);
  dc_vid_flip_fill_renderer (draw_type_fullscreen);
}


static const char *save_filename = "dgb_r0.3";

static uint8
find_vmu ()
{
  int n;
  uint8 addr;

  n = 0;
  for (;;)
  {
    addr = dc_maple_find_nth_device (n, MAPLE_FUNC_MEMCARD);
    ++n;
    if (!addr)
      return 0;
    else
    {
      if (ndc_vmu_get_dirent (NULL, addr, save_filename) >= 0)
	    return addr;
    }
  }
  return 0;
}

#if 0
static void
load_option_from_vmu (uint8 addr)
{
  int32 firstblk = ndc_vmu_get_dirent (NULL, addr, save_filename);
  if (firstblk < 0) return;

  uint8 buf[1024];
  if (ndc_vmu_read_blocks (addr, firstblk, buf, 2) != 2)
  {
    dc_put_error ("error loading user settings");
    return;
  }

#if 0
  set_cont_scheme                         (buf[640 + 0]);
  set_sound                               (buf[640 + 1]);
  set_use_vmu                             (buf[640 + 2]);
  set_bi_filter                           (buf[640 + 3]);
#endif
}
#endif

static const uint8 gnuboydc_icon[] = {
#include "dc/icons/gnuboydc-icon.h"
};

static void
save_option_to_vmu (uint8 addr)
{

  if (!addr) return;

  dc_print ("saving user settings...");

  uint8 buf[1024];
  ndc_vmu_create_vmu_header (buf, save_filename, "gnuboy/DC settings compat0.3",
			     1024 - 128, gnuboydc_icon);
  memset(buf + 640, 0, 1024 - 640);
#if 0
  buf[640 + 0] = cont_scheme;
  buf[640 + 1] = sound_on;
  buf[640 + 2] = use_vmu;
  buf[640 + 3] = bilinear_filter;
  buf[640 + 4] = fm_config.vertical_wait;
  buf[640 + 5] = fm_config.horizontal_wait;
#endif
  ndc_vmu_do_crc (buf, 1024);

  int32 firstblk = ndc_vmu_get_dirent (NULL, addr, save_filename);
  if (firstblk < 0)
  {
    uint8 free_blocks = ndc_vmu_check_free_blocks (NULL, addr);

    if (free_blocks < 2)
    {
      dc_put_error ("You need 2 free blocks to Save user settings");
      return;
    }

    firstblk = ndc_vmu_allocate_file (addr, save_filename, 2);
    if (firstblk < 0)
    {
      dc_put_error ("error updating fat & directory");
      return;
    }
  }

  if (ndc_vmu_write_blocks (addr, firstblk, buf, 2) != 2)
  {
    dc_put_error ("error writing data");
    return;
  }
}

KOS_INIT_FLAGS(INIT_DEFAULT);

int main()
{

  dc_vid_init ();

  dc_maple_controller_init ();

  vmu_set_icon (vmu_kos_xpm);

  load_bmp (startup_image, DC_FILES_ROOTPATH "/pics/startup.bmp");

  draw_startup_image ("startup");
  load_bmp (menu_image, DC_FILES_ROOTPATH "/pics/menu.bmp");
  draw_startup_image ("startup.");
  load_bmp (credits_image, DC_FILES_ROOTPATH "/pics/credits.bmp");
  draw_startup_image ("startup..");
  load_bmp (options_image, DC_FILES_ROOTPATH "/pics/options.bmp");
  draw_startup_image ("startup...");
  init_menus ();

  sprintf (last_romfile, DC_FILES_ROOTPATH "/gb/default.gb");

  draw_startup_image ("load setting");

  dc_mvmu = find_vmu ();
  if (!dc_mvmu)
  {
    dc_mvmu = do_vmu_select_menu ();
    if (dc_mvmu)
      save_option_to_vmu (dc_mvmu);
  }

  run_game (DC_FILES_ROOTPATH "/gb/default.gb");

  load_main_menu ();

  {
    char vmu_screen[48 * 32];
    vmu_icon_clear (vmu_screen, 1);
    vmu_set_icon(vmu_screen);
  }

  arch_return();
  return 0;
}

